# -----------------------------------------------------------------------------
# sheng: context.py
#
# Copyright (c) 2021
# luojiahai
# All rights reserved.
#
# Latest version: https://github.com/luojiahai/sheng
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -----------------------------------------------------------------------------


# -----------------------------------------------------------------------------
# BEGIN context.py
# -----------------------------------------------------------------------------

from .exception import *


class Context(object):

    def __init__(self, symtable):
        self._symtable = symtable

    @property
    def symtable(self):
        return self._symtable


class SymbolTable(object):

    def __init__(self, parent=None):
        self._symbols = dict()
        if (parent):
            for _, symbol in parent.symbols.items():
                self.insert(symbol.identifier, symbol.node)

    @property
    def symbols(self):
        return self._symbols
    
    def insert(self, identifier, node):
        symbol = Symbol(identifier, node)
        self.symbols[identifier] = symbol

    def lookup(self, identifier):
        if (identifier not in self.symbols):
            return None
        return self.symbols[identifier]


class Symbol(object):

    def __init__(self, identifier, node):
        self._identifier = identifier
        self._node = node

    @property
    def identifier(self):
        return self._identifier

    @property
    def node(self):
        return self._node


# -----------------------------------------------------------------------------
# END context.py
# -----------------------------------------------------------------------------



# -----------------------------------------------------------------------------
# sheng: parse.py
#
# Copyright (c) 2021
# luojiahai
# All rights reserved.
#
# Latest version: https://github.com/luojiahai/sheng
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -----------------------------------------------------------------------------


# -----------------------------------------------------------------------------
# BEGIN parse.py
# -----------------------------------------------------------------------------

from .ply import yacc
from .lex import *
from .ast import *
from .builtins import *
from .exception import *


# Token list
tokens = Lexer().tokens

# Precedence rules
precedence = (
    # ('left', 'OR',),
    # ('left', 'AND',),
    # ('right', 'NOT',),
    ('left', 'EQEQUAL', 'NOTEQUAL', 'LESS', 'GREATER', 'LESSEQUAL', 'GREATEREQUAL',),
    ('left', 'PLUS', 'MINUS',),
    # ('left', 'STAR', 'SLASH', 'DOUBLESLASH', 'PERCENT',),
    # ('right', 'UADD', 'USUB',),
    # ('right', 'DOUBLESTAR',),
)

# Rule for a program
def p_program(p):
    '''program : block
               | empty'''
    block = []
    if (p[1]): block = p[1]
    p[0] = Program(block)

# Rule for block
def p_block(p):
    '''block : statements'''
    p[0] = p[1]

# Rule for statements
def p_statements(p):
    '''statements : statements statement
                  | statement'''
    if (len(p) == 2 and p[1]):
        statement = p[1]
        p[0] = [statement]
    elif (len(p) == 3):
        statements = p[1]
        if (p[2]): statements.append(p[2])
        p[0] = statements

# Rule for statement
def p_statement(p):
    '''statement : statement_assignment
                 | statement_return
                 | statement_pass
                 | statement_assert
                 | statement_break
                 | statement_continue
                 | statement_function_def
                 | statement_if
                 | statement_class_def
                 | statement_iterate
                 | statement_loop
                 | expression'''
    p[0] = p[1]

# Statement rule for assignment
def p_statement_assignment(p):
    '''statement_assignment : NAME EQUAL expression'''
    name = Name(p[1])
    p[0] = Assignment(name, p[3])

# Statement rule for return
def p_statement_return(p):
    '''statement_return : RETURN expression
                        | RETURN'''
    if (len(p) == 2):
        p[0] = Return(None)
    elif (len(p) == 3):
        p[0] = Return(p[2])

# Statement rule for pass
def p_statement_pass(p):
    '''statement_pass : PASS'''
    p[0] = Pass(p[1])

# Statement rule for assert
def p_statement_assert(p):
    '''statement_assert : ASSERT expression'''
    p[0] = Assert(p[2])

# Statement rule for break
def p_statement_break(p):
    '''statement_break : BREAK'''
    p[0] = Break(p[1])

# Statement rule for continue
def p_statement_continue(p):
    '''statement_continue : CONTINUE'''
    p[0] = Continue(p[1])

# Statement rule for function definition
def p_statement_function_def(p):
    '''statement_function_def : FUNCTIONDEF NAME LPAR parameters RPAR BEGIN block END
                              | FUNCTIONDEF NAME LPAR RPAR BEGIN block END'''
    if (len(p) == 8):
        name = Name(p[2])
        body = p[6]
        p[0] = FunctionDef(name, [], body)
    elif (len(p) == 9):
        name = Name(p[2])
        parameters = p[4]
        body = p[7]
        p[0] = FunctionDef(name, parameters, body)

# Statement rule for if
def p_statement_if(p):
    '''statement_if : IF expression THEN block statement_elif END'''
    expression = p[2]
    block = p[4]
    statement_elif = p[5]
    p[0] = If(expression, block, statement_elif=statement_elif)

def p_statement_if_else(p):
    '''statement_if : IF expression THEN block block_else END
                    | IF expression THEN block END'''
    if (len(p) == 6):
        expression = p[2]
        block = p[4]
        p[0] = If(expression, block)
    elif (len(p) == 7):
        expression = p[2]
        block = p[4]
        block_else = p[5]
        p[0] = If(expression, block, block_else=block_else)

# Statement rule for elif
def p_statement_elif(p):
    '''statement_elif : ELIF expression THEN block statement_elif'''
    expression = p[2]
    block = p[4]
    statement_elif = p[5]
    p[0] = [Elif(expression, block)] + statement_elif

def p_statement_elif_else(p):
    '''statement_elif : ELIF expression THEN block block_else
                      | ELIF expression THEN block'''
    if (len(p) == 5):
        expression = p[2]
        block = p[4]
        p[0] = [Elif(expression, block)]
    elif (len(p) == 6):
        expression = p[2]
        block = p[4]
        block_else = p[5]
        p[0] = [Elif(expression, block, block_else=block_else)]

# Rule for else block
def p_block_else(p):
    '''block_else : ELSE block'''
    p[0] = p[2]

# Statement rule for class definition
def p_statement_class_def(p):
    '''statement_class_def : CLASSDEF NAME BEGIN block END'''
    name = Name(p[2])
    block = p[4]
    p[0] = ClassDef(name, block=block)

# Statement rule for iterate
def p_statement_iterate(p):
    '''statement_iterate : ITERATE expression FOR NAME BEGIN block END'''
    iterable = p[2]
    variable = Name(p[4])
    statements = p[6]
    p[0] = Iterate(iterable, variable, statements)

# Statement rule for loop
def p_statement_loop(p):
    '''statement_loop : LOOP expression BEGIN block END'''
    expression = p[2]
    statements = p[4]
    p[0] = Loop(expression, statements)

# Rule for expressions
def p_expressions(p):
    '''expressions : expression comma_expressions COMMA
                   | expression comma_expressions
                   | expression COMMA
                   | expression'''
    if ((len(p) == 3 and p[2] == ',') or len(p) == 2):
        expression = p[1]
        p[0] = [expression]
    elif (len(p) == 4 or len(p) == 3):
        comma_expressions = p[2]
        if (p[1]): comma_expressions.insert(0, p[1])
        p[0] = comma_expressions

# Rule for expressions: (',' expression)+
def p_comma_expressions(p):
    '''comma_expressions : comma_expressions COMMA expression
                         | COMMA expression'''
    if (len(p) == 3):
        expression = p[2]
        p[0] = [expression]
    elif (len(p) == 4):
        comma_expressions = p[1]
        if (p[3]): comma_expressions.append(p[3])
        p[0] = comma_expressions

# Rule for parameters
def p_parameters(p):
    '''parameters : parameter parameters_comma COMMA
                  | parameter parameters_comma
                  | parameter COMMA
                  | parameter'''
    if ((len(p) == 3 and p[2] == ',') or len(p) == 2):
        parameter = p[1]
        p[0] = [parameter]
    elif (len(p) == 4 or len(p) == 3):
        parameters_comma = p[2]
        if (p[1]): parameters_comma.insert(0, p[1])
        p[0] = parameters_comma

# Rule for parameter with no default
def p_parameters_comma(p):
    '''parameters_comma : parameters_comma COMMA parameter
                        | COMMA parameter'''
    if (len(p) == 3):
        paramter = p[2]
        p[0] = [paramter]
    elif (len(p) == 4):
        parameters_comma = p[1]
        if (p[3]): parameters_comma.append(p[3])
        p[0] = parameters_comma

# Rule for parameter
def p_parameter(p):
    '''parameter : NAME'''
    p[0] = Name(p[1])

# Expression rule for group expression
def p_expression_group(p):
    '''expression : LPAR expression RPAR'''
    p[0] = p[2]

# Expression rule for disjunction
def p_expression_disjunction(p):
    '''expression : disjunction'''
    p[0] = p[1]

# Rule for disjunction
def p_disjunction(p):
    '''disjunction : conjunction or_conjunctions
                   | conjunction'''
    if (len(p) == 2):
        p[0] = Disjunction([p[1]])
    elif (len(p) == 3):
        p[0] = Disjunction([p[1]] + p[2])

# Rule for (OR conjunction)+
def p_or_conjunctions(p):
    '''or_conjunctions : or_conjunctions OR conjunction
                       | OR conjunction'''
    if (len(p) == 3):
        p[0] = [p[2]]
    elif (len(p) == 4):
        or_conjunctions = p[1]
        or_conjunctions.append(p[3])
        p[0] = or_conjunctions

# Rule for conjunction
def p_conjunction(p):
    '''conjunction : inversion and_inversions
                   | inversion'''
    if (len(p) == 2):
        p[0] = Conjunction([p[1]])
    elif (len(p) == 3):
        p[0] = Conjunction([p[1]] + p[2])

# Rule for (AND inversion)+
def p_and_inversions(p):
    '''and_inversions : and_inversions AND inversion
                      | AND inversion'''
    if (len(p) == 3):
        p[0] = [p[2]]
    elif (len(p) == 4):
        and_inversions = p[1]
        and_inversions.append(p[3])
        p[0] = and_inversions

# Rule for inversion
def p_inversion(p):
    '''inversion : NOT inversion
                 | comparison'''
    if (len(p) == 2):
        p[0] = p[1]
    elif (len(p) == 3):
        p[0] = Inversion(p[2])

# Rule for comparison
def p_comparison(p):
    '''comparison : sum EQEQUAL sum
                  | sum NOTEQUAL sum
                  | sum LESS sum
                  | sum LESSEQUAL sum
                  | sum GREATER sum
                  | sum GREATEREQUAL sum
                  | sum'''
    if (len(p) == 2):
        p[0] = p[1]
    else:
        left = p[1]
        right = p[3]
        operator = EXACT_TOKEN_TYPES[p[2]]
        if (operator == 'EQEQUAL'):
            p[0] = Equal(left, right)
        elif (operator == 'NOTEQUAL'):
            p[0] = NotEqual(left, right)
        elif (operator == 'LESS'):
            p[0] = Less(left, right)
        elif (operator == 'LESSEQUAL'):
            p[0] = LessEqual(left, right)
        elif (operator == 'GREATER'):
            p[0] = Greater(left, right)
        elif (operator == 'GREATEREQUAL'):
            p[0] = GreaterEqual(left, right)

# Rule for sum
def p_sum(p):
    '''sum : sum PLUS term
           | sum MINUS term
           | term'''
    if (len(p) == 2):
        p[0] = p[1]
    else:
        left = p[1]
        right = p[3]
        operator = EXACT_TOKEN_TYPES[p[2]]
        if (operator == 'PLUS'):
            p[0] = Add(left, right)
        elif (operator == 'MINUS'):
            p[0] = Subtract(left, right)

# Rule for term
def p_term(p):
    '''term : term STAR factor
            | term SLASH factor
            | term DOUBLESLASH factor
            | term PERCENT factor
            | factor'''
    if (len(p) == 2):
        p[0] = p[1]
    else:
        left = p[1]
        right = p[3]
        operator = EXACT_TOKEN_TYPES[p[2]]
        if (operator == 'STAR'):
            p[0] = Multiply(left, right)
        elif (operator == 'SLASH'):
            p[0] = Divide(left, right)
        elif (operator == 'DOUBLESLASH'):
            p[0] = FloorDivide(left, right)
        elif (operator == 'PERCENT'):
            p[0] = Modulo(left, right)

# Rule for factor
def p_factor(p):
    '''factor : PLUS factor
              | MINUS factor
              | power'''
    if (len(p) == 2):
        p[0] = p[1]
    else:
        value = p[2]
        operator = EXACT_TOKEN_TYPES[p[1]]
        if (operator == 'PLUS'):
            p[0] = UAdd(value)
        elif (operator == 'MINUS'):
            p[0] = USubtract(value)

# Rule for power
def p_power(p):
    '''power : atom DOUBLESTAR factor
             | atom'''
    if (len(p) == 2):
        p[0] = p[1]
    else:
        p[0] = Power(p[1], p[3])

# Atom rule for name
def p_atom_name(p):
    '''atom : NAME'''
    p[0] = Name(p[1])

# Atom rule for boolean
def p_atom_boolean(p):
    '''atom : TRUE
            | FALSE'''
    atom = EXACT_TOKEN_TYPES[p[1]]
    if (atom == 'TRUE'):
        p[0] = Boolean(1)
    elif (atom == 'FALSE'):
        p[0] = Boolean(0)

# Atom rule for null
def p_atom_null(p):
    '''atom : NULL'''
    p[0] = Null(None)

# Atom rule for string
def p_atom_string(p):
    '''atom : STRING'''
    p[0] = String(p[1])

# Atom rule for integer number
def p_atom_integer(p):
    '''atom : INTEGER'''
    p[0] = Integer(p[1])

# Atom rule for float number
def p_atom_float(p):
    '''atom : FLOAT'''
    p[0] = Float(p[1])

# Atom rule for complex number
def p_atom_complex(p):
    '''atom : COMPLEX'''
    p[0] = Complex(p[1])

# Atom rule for tuple
def p_atom_tuple(p):
    '''atom : LPAR expressions RPAR
            | LPAR RPAR'''
    if (len(p) == 3):
        p[0] = Tuple([])
    elif (len(p) == 4):
        expressions = p[2]
        p[0] = Tuple(expressions)

# Atom rule for list
def p_atom_list(p):
    '''atom : LSQB expressions RSQB
            | LSQB RSQB'''
    if (len(p) == 3):
        p[0] = List([])
    elif (len(p) == 4):
        expressions = p[2]
        p[0] = List(expressions)

# Atom rule for dict
def p_atom_dict(p):
    '''atom : LBRACE key_value_pairs RBRACE
            | LBRACE RBRACE'''
    if (len(p) == 3):
        p[0] = Dict([])
    elif (len(p) == 4):
        key_value_pairs = p[2]
        p[0] = Dict(key_value_pairs)

# Rule for key-value pairs
def p_key_value_pairs(p):
    '''key_value_pairs : key_value_pairs COMMA key_value_pair
                       | key_value_pair'''
    if (len(p) == 2 and p[1]):
        key_value_pair = p[1]
        p[0] = [key_value_pair]
    elif (len(p) == 4):
        key_value_pairs = p[1]
        if (p[3]): key_value_pairs.append(p[3])
        p[0] = key_value_pairs

# Rule for key-value pair
def p_key_value_pair(p):
    '''key_value_pair : expression COLON expression'''
    expression_key = p[1]
    expression_value = p[3]
    p[0] = KeyValuePair(expression_key, expression_value)

# Atom rule for iterable index
def p_atom_iterable_index(p):
    '''atom : NAME INDEX expression'''
    name = Name(p[1])
    expression = p[3]
    p[0] = IterableIndex(name, expression)

# Atom rule for function call
def p_atom_functioncall(p):
    '''atom : function_call'''
    p[0] = p[1]

# Rule for function call
def p_functioncall(p):
    '''function_call : NAME LPAR expressions RPAR
                     | NAME LPAR RPAR'''
    if (len(p) == 4):
        name = Name(p[1])
        p[0] = FunctionCall(name, [])
    elif (len(p) == 5):
        name = Name(p[1])
        expressions = p[3]
        p[0] = FunctionCall(name, expressions)

# Atom rule for class member call
def p_atom_classmembercall(p):
    '''atom : class_member_call'''
    name = p[1][0]
    methods = p[1][1:]
    p[0] = ClassMemberCall(name, methods)

# Rule for class member call
def p_classmembercall(p):
    '''class_member_call : class_member_call DOT function_call
                         | class_member_call DOT NAME
                         | NAME DOT function_call'''
    if (len(p) == 4 and type(p[1]) == str):
        p[0] = [Name(p[1]), p[3]]
    elif (len(p) == 4):
        class_member_call = p[1]
        if (p[3]):
            if (type(p[3]) == str): 
                class_member_call.append(Name(p[3]))
            else:
                class_member_call.append(p[3])
        p[0] = class_member_call

# Rule for nothing
def p_empty(p):
    '''empty : '''
    pass

# Error rule for syntax error
def p_error(p):
    if (p):
        raise SyntaxException(f'无效的标记 \'{p.value}\'，第 {p.lineno} 行')
    else:
        raise SyntaxException(f'文件结尾')


# -----------------------------------------------------------------------------
# END parse.py
# -----------------------------------------------------------------------------



# -----------------------------------------------------------------------------
# sheng: compile.py
#
# Copyright (c) 2021
# luojiahai
# All rights reserved.
#
# Latest version: https://github.com/luojiahai/sheng
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -----------------------------------------------------------------------------


# -----------------------------------------------------------------------------
# BEGIN compile.py
# -----------------------------------------------------------------------------

import logging
import inspect
import traceback
from .lex import *
from .parse import *
from .ast import *
from .context import *
from .builtins import *


def execute(data):
    try:
        debug = True if (Global().is_debug) else False
        
        # Set up a logging object
        log = None
        if (debug):
            logging.basicConfig(
                level = logging.DEBUG,
                stream = Global().log_stream,
                filemode = 'a',
                format = '「日志」【排除故障】 %(filename)s:%(lineno)d:%(message)s'
            )
            log = logging.getLogger()

        # Lex
        lexer = Lexer()
        lexer.build(debug=debug, debuglog=log)
        
        # Initialise context with symbol table
        symtable = SymbolTable()
        context = Context(symtable)

        # Add builtin functions to the symbol table
        def get_default_args(func):
            signature = inspect.signature(func)
            return {
                k: v.default
                for k, v in signature.parameters.items()
                if v.default is not inspect.Parameter.empty
            }
        for identifier, function_pointer in BUILTIN_FUNCTIONS.items():
            name = Name(identifier)
            parameters = [
                Name(arg)
                for arg in inspect.getfullargspec(function_pointer)[0]
            ]
            default_args = get_default_args(function_pointer)
            functiondef = FunctionDef(
                name, parameters, function_pointer, default_args
            )
            context.symtable.insert(identifier, functiondef)

        # Parse
        parser = yacc.yacc(debug=debug, debuglog=log)
        ast = parser.parse(data)
        return ast.evaluate(context)
        
    except BaseException as e:
        if (isinstance(e, ShengException)):
            Global().log(e)
            if (Global().is_debug):
                Global().log(f'「日志」【排除故障】 {traceback.format_exc()}')
        else:
            error_message = f'「编译器报错（未被捕捉的错误）」【{type(e).__name__}】\n' \
                '(1) 请重新运行命令，在命令中添加\'--debug\'获取运行日志\n' \
                '(2) 在\'./log\'文件夹中查看运行日志文件\n' \
                '(3) 提交至结绳的GitHub仓库请求修复'
            Global().log(error_message)
            if (Global().is_debug):
                Global().log(f'「日志」【排除故障】 {traceback.format_exc()}')

    finally:
        return None


# -----------------------------------------------------------------------------
# ENG compile.py
# -----------------------------------------------------------------------------



# -----------------------------------------------------------------------------
# sheng: builtins.py
#
# Copyright (c) 2021
# luojiahai
# All rights reserved.
#
# Latest version: https://github.com/luojiahai/sheng
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -----------------------------------------------------------------------------


# -----------------------------------------------------------------------------
# BEGIN builtins.py
# -----------------------------------------------------------------------------

from .utils import *
from .context import *


# -----------------------------------------------------------------------------
# Build-in types
# -----------------------------------------------------------------------------

class BaseType(object):

    def __init__(self, type_name):
        self._type = type_name

    @property
    def type(self):
        return f'<类型 \'{self._type}\'>'


class NullType(BaseType):

    def __init__(self, value):
        BaseType.__init__(self, '空类型')
        self.value = value

    def __repr__(self):
        return '空'

    def __str__(self):
        return '空'

    def __bool__(self):
        return bool(None)

    def __eq__(self, other):
        return other == None

    def __ne__(self, other):
        return other != None


class BooleanType(BaseType, int):

    def __init__(self, value):
        BaseType.__init__(self, '布林')
        self.value = value

    def __repr__(self):
        return '对' if self.value else '错'

    def __str__(self):
        return '对' if self.value else '错'

    def __bool__(self):
        return self.value


class IntegerType(BaseType, int):

    def __init__(self, value):
        BaseType.__init__(self, '整数')
        self.value = value


class FloatType(BaseType, float):

    def __init__(self, value):
        BaseType.__init__(self, '浮点数')
        self.value = value


class ComplexType(BaseType, complex):

    def __init__(self, value):
        BaseType.__init__(self, '复数')
        self.value = value


class StringType(BaseType, str):

    def __init__(self, value):
        BaseType.__init__(self, '字符串')
        self.value = value


class TupleType(BaseType, tuple):

    def __init__(self, value):
        BaseType.__init__(self, '元组')
        self.value = value
    
    def __new__(self, value):
        return tuple.__new__(self, value)


class ListType(BaseType, list):

    def __init__(self, value):
        BaseType.__init__(self, '列表')
        self.extend(value)


class DictType(BaseType, dict):

    def __init__(self, value):
        BaseType.__init__(self, '字典')
        self.update(value)


class ClassType(BaseType):

    def __init__(self, class_name):
        BaseType.__init__(self, class_name)
        symtable = SymbolTable()
        self._context = Context(symtable)
    
    @property
    def context(self):
        return self._context


BUILTIN_TYPES = {
    'NoneType': NullType,
    'bool': BooleanType,
    'int': IntegerType,
    'float': FloatType,
    'complex': ComplexType,
    'str': StringType,
    'tuple': TupleType,
    'list': ListType,
    'dict': DictType,
}


# -----------------------------------------------------------------------------
# Built-in functions
# -----------------------------------------------------------------------------

def builtin_print(value=''):
    print(value, file=Global().output_stream)
    if (Global().is_debug):
        Global().log(f'「日志」【排除故障】【{Global().output_stream.name}】 {value}')
    return None

def builtin_input(prompt=''):
    value = input(prompt)
    if (Global().is_debug):
        Global().log(f'「日志」【排除故障】【<stdin>】 {prompt}{value}')
    return value

def builtin_abs(value):
    return abs(value)

def builtin_max(iterable):
    return max(iterable)

def builtin_min(iterable):
    return min(iterable)

def builtin_type(value):
    return value.type


BUILTIN_FUNCTIONS = {
    '打印': builtin_print,
    '输入': builtin_input,
    '绝对值': builtin_abs,
    '最大值': builtin_max,
    '最小值': builtin_min,
    '类型': builtin_type,
}


# -----------------------------------------------------------------------------
# END builtins.py
# -----------------------------------------------------------------------------



# -----------------------------------------------------------------------------
# sheng: ast.py
#
# Copyright (c) 2021
# luojiahai
# All rights reserved.
#
# Latest version: https://github.com/luojiahai/sheng
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -----------------------------------------------------------------------------


# -----------------------------------------------------------------------------
# BEGIN ast.py
# -----------------------------------------------------------------------------

from abc import *
from .context import *
from .exception import *
from .builtins import *


# -----------------------------------------------------------------------------
# Generic node class for Abstract Syntax Tree
# -----------------------------------------------------------------------------

class ASTNode(ABC):

    def evaluate(self, context):
        raise NotImplementedError()


# -----------------------------------------------------------------------------
# Utility class for Abstract Syntax Tree
# -----------------------------------------------------------------------------

class ASTUtil(object):

    def evaluate_base_type(value):
        type_name = type(value).__name__
        base_types = [e.__name__ for e in BUILTIN_TYPES.values()]
        if (type_name in base_types):
            return value
        elif (type_name in BUILTIN_TYPES):
            base_type = BUILTIN_TYPES[type_name]
            return base_type(value)
        # else:
        #     raise RuntimeException(f'无效的操作数类型 {type_name}')

    def evaluate_block(context, sequence):
        for node in sequence:
            value = node.evaluate(context)
            if (value):
                return value
        return None

    def instantiate_class(context, classdef):
        instance = ClassType(classdef.name.identifier)
        properties = classdef.properties
        methods = classdef.methods

        # Insert parent symbols to class context
        for _, s in context.symtable.symbols.items():
            instance.context.symtable.insert(s.identifier, s.node)

        # Evaluate assignment statements
        for _, statement in properties.items():
            statement.evaluate(instance.context)

        # Evaluate function def statements
        for _, statement in methods.items():
            statement.evaluate(instance.context)
        
        # Check if a constructor defined, evaluate if so
        if ('构造' in methods):
            functiondef = methods['构造']
            functioncall = FunctionCall(Name('构造'), functiondef.parameters)
            functioncall.evaluate(instance.context)
            
        return instance


# -----------------------------------------------------------------------------
# Program
# -----------------------------------------------------------------------------

class Program(ASTNode):

    def __init__(self, block):
        self._block = block

    @property
    def block(self):
        return self._block

    def evaluate(self, context):
        return ASTUtil.evaluate_block(
            context, self.block
        )


# -----------------------------------------------------------------------------
# Statement
# -----------------------------------------------------------------------------

class Statement(ASTNode): pass


# -----------------------------------------------------------------------------
# Assignment statement
# -----------------------------------------------------------------------------

class Assignment(Statement):

    def __init__(self, name, node):
        self._name = name
        self._node = node

    @property
    def name(self):
        return self._name

    @property
    def node(self):
        return self._node

    def evaluate(self, context):
        identifier = self.name.identifier
        node = Atom(self.node.evaluate(context))
        context.symtable.insert(identifier, node)
        return None


# -----------------------------------------------------------------------------
# Pass statement
# -----------------------------------------------------------------------------

class Pass(Statement):

    def __init__(self, value):
        self._value = value

    @property
    def value(self):
        return self._value

    def evaluate(self, context):
        pass


# -----------------------------------------------------------------------------
# Assert statement
# -----------------------------------------------------------------------------

class Assert(Statement):

    def __init__(self, expression):
        self._expression = expression

    @property
    def expression(self):
        return self._expression

    def evaluate(self, context):
        if (not self.expression.evaluate(context)):
            raise AssertionException()
        return None


# -----------------------------------------------------------------------------
# Return statement
# -----------------------------------------------------------------------------

class Return(Statement):

    def __init__(self, expression):
        self._expression = expression

    @property
    def expression(self):
        return self._expression

    def evaluate(self, context):
        value = self.expression.evaluate(context)
        ReturnValue = type('ReturnValue', (type(value), ), {})
        return ReturnValue(value)


# -----------------------------------------------------------------------------
# Break statement
# -----------------------------------------------------------------------------

class Break(Statement):

    def __init__(self, value):
        self._value = value

    @property
    def value(self):
        return self._value

    def evaluate(self, context):
        BreakValue = type('BreakValue', (object, ), {})
        return BreakValue()


# -----------------------------------------------------------------------------
# Continue statement
# -----------------------------------------------------------------------------

class Continue(Statement):

    def __init__(self, value):
        self._value = value

    @property
    def value(self):
        return self._value

    def evaluate(self, context):
        ContinueValue = type('ContinueValue', (object, ), {})
        return ContinueValue()


# -----------------------------------------------------------------------------
# Function definition statement
# -----------------------------------------------------------------------------

class FunctionDef(Statement, BaseType):

    def __init__(self, name, parameters, body, default_args=[]):
        BaseType.__init__(self, '函数')
        self._name = name
        self._parameters = parameters
        self._body = body
        self._default_args = default_args

    @property
    def name(self):
        return self._name

    @property
    def parameters(self):
        return self._parameters

    @property
    def body(self):
        return self._body

    @property
    def default_args(self):
        return self._default_args

    def evaluate(self, context):
        identifier = self.name.identifier
        context.symtable.insert(identifier, self)
        return None


# -----------------------------------------------------------------------------
# If and Elif statements
# -----------------------------------------------------------------------------

class If(Statement):

    def __init__(self, expression, statements, 
                 statement_elif=[], block_else=None):
        self._expression = expression
        self._statements = statements
        self._statement_elif = statement_elif
        self._block_else = block_else

    @property
    def expression(self):
        return self._expression
    
    @property
    def statements(self):
        return self._statements

    @property
    def statement_elif(self):
        return self._statement_elif

    @property
    def block_else(self):
        return self._block_else

    def evaluate(self, context):
        if (self.expression.evaluate(context)):
            return ASTUtil.evaluate_block(
                context, self.statements
            )
        else:
            # If there exists an elif statement
            for statement_elif in self.statement_elif:
                if (statement_elif.is_true(context)):
                    return statement_elif.evaluate(context)
                elif (statement_elif.block_else):
                    return ASTUtil.evaluate_block(
                        context, statement_elif.block_else
                    )
            
            # If there exists an else block
            if (self.block_else):
                return ASTUtil.evaluate_block(
                    context, self.block_else
                )
            return None


class Elif(Statement):

    def __init__(self, expression, statements, block_else=None):
        self._expression = expression
        self._statements = statements
        self._block_else = block_else

    @property
    def expression(self):
        return self._expression
    
    @property
    def statements(self):
        return self._statements

    @property
    def block_else(self):
        return self._block_else

    def is_true(self, context):
        return self.expression.evaluate(context)

    def evaluate(self, context):
        for node in self.statements:
            value = node.evaluate(context)
            if (value):
                return value
        return None


# -----------------------------------------------------------------------------
# Class definition statement
# -----------------------------------------------------------------------------

class ClassDef(Statement, BaseType):

    def __init__(self, name, block=[]):
        BaseType.__init__(self, '类')
        self._name = name
        properties = dict()
        methods = dict()
        for statement in block:
            if (type(statement) == Assignment):
                properties[statement.name.identifier] = statement
            elif (type(statement) == FunctionDef):
                methods[statement.name.identifier] = statement
        self._properties = properties
        self._methods = methods
    
    @property
    def name(self):
        return self._name

    @property
    def properties(self):
        return self._properties

    @property
    def methods(self):
        return self._methods

    def evaluate(self, context):
        identifier = self.name.identifier
        context.symtable.insert(identifier, self)
        return None


# -----------------------------------------------------------------------------
# Iterate statement
# -----------------------------------------------------------------------------

class Iterate(Statement):

    def __init__(self, iterable, variable, statements):
        self._iterable = iterable
        self._variable = variable
        self._statements = statements

    @property
    def iterable(self):
        return self._iterable
    
    @property
    def variable(self):
        return self._variable

    @property
    def statements(self):
        return self._statements

    def evaluate(self, context):
        variable_identifier = self.variable.identifier
        for e in self.iterable.evaluate(context):
            node = Atom(e)
            context.symtable.insert(variable_identifier, node)
            value = ASTUtil.evaluate_block(
                context, self.statements
            )
            if (type(value).__name__ == 'ReturnValue'):
                return value
            if (type(value).__name__ == 'ContinueValue'):
                continue
            if (type(value).__name__ == 'BreakValue'):
                break
        return None


# -----------------------------------------------------------------------------
# Loop statement
# -----------------------------------------------------------------------------

class Loop(Statement):

    def __init__(self, expression, statements):
        self._expression = expression
        self._statements = statements

    @property
    def expression(self):
        return self._expression
    
    @property
    def statements(self):
        return self._statements

    def evaluate(self, context):
        while (self.expression.evaluate(context)):
            value = ASTUtil.evaluate_block(
                context, self.statements
            )
            if (type(value).__name__ == 'ReturnValue'):
                return value
            if (type(value).__name__ == 'ContinueValue'):
                continue
            if (type(value).__name__ == 'BreakValue'):
                break
        return None


# -----------------------------------------------------------------------------
# Expression
# -----------------------------------------------------------------------------

class Expression(ASTNode): pass


# -----------------------------------------------------------------------------
# Boolean operator expressions
# -----------------------------------------------------------------------------

class Disjunction(Expression):

    def __init__(self, conjunctions):
        self._conjunctions = conjunctions
    
    @property
    def conjunctions(self):
        return self._conjunctions

    def evaluate(self, context):
        if (len(self.conjunctions) == 1):
            return self.conjunctions[0].evaluate(context)
        else:
            value = self.conjunctions[0].evaluate(context)
            for conjunction in self.conjunctions[1:]:
                if (type(conjunction) != Inversion):
                    value = value or conjunction.evaluate(context)
            return ASTUtil.evaluate_base_type(value)


class Conjunction(Expression):

    def __init__(self, inversions):
        self._inversions = inversions
    
    @property
    def inversions(self):
        return self._inversions

    def evaluate(self, context):
        if (len(self.inversions) == 1):
            return self.inversions[0].evaluate(context)
        else:
            value = self.inversions[0].evaluate(context)
            for inversion in self.inversions[1:]:
                if (type(inversion) != Inversion):
                    value = value and inversion.evaluate(context)
            return ASTUtil.evaluate_base_type(value)


class Inversion(Expression):

    def __init__(self, inversion):
        self._inversion = inversion

    @property
    def inversion(self):
        return self._inversion

    def traverse(self, inversion, context):
        if (type(inversion) != Inversion):
            return not inversion.evaluate(context)
        else:
            return not self.traverse(inversion.inversion, context)
    
    def evaluate(self, context):
        return ASTUtil.evaluate_base_type(
            self.traverse(self.inversion, context)
        )


# -----------------------------------------------------------------------------
# Unary operator expressions
# -----------------------------------------------------------------------------

class UnaryOp(Expression):

    def __init__(self, node):
        self._node = node

    @property
    def node(self):
        return self._node


class UAdd(UnaryOp):

    def evaluate(self, context):
        value = self.node.evaluate(context)
        return ASTUtil.evaluate_base_type(+value)


class USubtract(UnaryOp):

    def evaluate(self, context):
        value = self.node.evaluate(context)
        return ASTUtil.evaluate_base_type(-value)


# -----------------------------------------------------------------------------
# Binary operator expressions
# -----------------------------------------------------------------------------

class BinaryOp(Expression):

    def __init__(self, left, right):
        self._left = left
        self._right = right

    @property
    def left(self):
        return self._left

    @property
    def right(self):
        return self._right


class Add(BinaryOp):

    def evaluate(self, context):
        lvalue = self.left.evaluate(context)
        rvalue = self.right.evaluate(context)
        return ASTUtil.evaluate_base_type(lvalue + rvalue)


class Subtract(BinaryOp):

    def evaluate(self, context):
        lvalue = self.left.evaluate(context)
        rvalue = self.right.evaluate(context)
        return ASTUtil.evaluate_base_type(lvalue - rvalue)


class Multiply(BinaryOp):

    def evaluate(self, context):
        lvalue = self.left.evaluate(context)
        rvalue = self.right.evaluate(context)
        return ASTUtil.evaluate_base_type(lvalue * rvalue)


class Divide(BinaryOp):

    def evaluate(self, context):
        lvalue = self.left.evaluate(context)
        rvalue = self.right.evaluate(context)
        return ASTUtil.evaluate_base_type(lvalue / rvalue)


class Modulo(BinaryOp):

    def evaluate(self, context):
        lvalue = self.left.evaluate(context)
        rvalue = self.right.evaluate(context)
        return ASTUtil.evaluate_base_type(lvalue % rvalue)


class Power(BinaryOp):

    def evaluate(self, context):
        lvalue = self.left.evaluate(context)
        rvalue = self.right.evaluate(context)
        return ASTUtil.evaluate_base_type(lvalue ** rvalue)


class FloorDivide(BinaryOp):

    def evaluate(self, context):
        lvalue = self.left.evaluate(context)
        rvalue = self.right.evaluate(context)
        return ASTUtil.evaluate_base_type(lvalue // rvalue)


# -----------------------------------------------------------------------------
# Compare operator expressions
# -----------------------------------------------------------------------------

class CompareOp(BinaryOp): pass


class Equal(CompareOp):

    def evaluate(self, context):
        lvalue = self.left.evaluate(context)
        rvalue = self.right.evaluate(context)
        return ASTUtil.evaluate_base_type(lvalue == rvalue)


class NotEqual(CompareOp):

    def evaluate(self, context):
        lvalue = self.left.evaluate(context)
        rvalue = self.right.evaluate(context)
        return ASTUtil.evaluate_base_type(lvalue != rvalue)


class Less(CompareOp):

    def evaluate(self, context):
        lvalue = self.left.evaluate(context)
        rvalue = self.right.evaluate(context)
        return ASTUtil.evaluate_base_type(lvalue < rvalue)


class LessEqual(CompareOp):

    def evaluate(self, context):
        lvalue = self.left.evaluate(context)
        rvalue = self.right.evaluate(context)
        return ASTUtil.evaluate_base_type(lvalue <= rvalue)


class Greater(CompareOp):

    def evaluate(self, context):
        lvalue = self.left.evaluate(context)
        rvalue = self.right.evaluate(context)
        return ASTUtil.evaluate_base_type(lvalue > rvalue)


class GreaterEqual(CompareOp):

    def evaluate(self, context):
        lvalue = self.left.evaluate(context)
        rvalue = self.right.evaluate(context)
        return ASTUtil.evaluate_base_type(lvalue >= rvalue)


# -----------------------------------------------------------------------------
# Atom
# -----------------------------------------------------------------------------

class Atom(Expression):

    def __init__(self, value):
        self._value = value

    @property
    def value(self):
        return self._value

    def evaluate(self, context):
        return self.value


class Name(Atom):

    def __init__(self, identifier):
        self._identifier = identifier

    @property
    def identifier(self):
        return self._identifier

    def evaluate(self, context):
        identifier = self._identifier
        symbol = context.symtable.lookup(identifier)
        if (not symbol):
            raise RuntimeException(f'{identifier} 未定义')
        node = symbol.node
        return node.evaluate(context)


class Boolean(Atom):

    def evaluate(self, context):
        return ASTUtil.evaluate_base_type(bool(self.value))


class Null(Atom):

    def evaluate(self, context):
        return ASTUtil.evaluate_base_type(None)


class String(Atom):

    def evaluate(self, context):
        return ASTUtil.evaluate_base_type(str(self.value[1:-1]))


class Integer(Atom):

    def evaluate(self, context):
        return ASTUtil.evaluate_base_type(int(self.value))


class Float(Atom):

    def evaluate(self, context):
        return ASTUtil.evaluate_base_type(float(self.value))


class Complex(Atom):

    def evaluate(self, context):
        return ASTUtil.evaluate_base_type(complex(self.value))


# -----------------------------------------------------------------------------
# Iterable
# -----------------------------------------------------------------------------

class Iterable(Atom): pass


# -----------------------------------------------------------------------------
# Tuple
# -----------------------------------------------------------------------------

class Tuple(Iterable):

    def __init__(self, expressions):
        self._expressions = expressions

    @property
    def expressions(self):
        return self._expressions
    
    def evaluate(self, context):
        return ASTUtil.evaluate_base_type(tuple([
            expression.evaluate(context) 
            for expression in self.expressions
        ]))


# -----------------------------------------------------------------------------
# List
# -----------------------------------------------------------------------------

class List(Iterable):

    def __init__(self, expressions):
        self._expressions = expressions

    @property
    def expressions(self):
        return self._expressions
    
    def evaluate(self, context):
        return ASTUtil.evaluate_base_type([
            expression.evaluate(context) 
            for expression in self.expressions
        ])


# -----------------------------------------------------------------------------
# Dictionary
# -----------------------------------------------------------------------------

class Dict(Iterable):

    def __init__(self, key_value_pairs):
        self._key_value_pairs = key_value_pairs

    @property
    def key_value_pairs(self):
        return self._key_value_pairs

    def evaluate(self, context):
        the_dict = dict()
        for key_value_pair in self.key_value_pairs:
            key, value = key_value_pair.evaluate(context)
            the_dict[key] = value
        return ASTUtil.evaluate_base_type(the_dict)


class KeyValuePair(Atom):
    
    def __init__(self, expression_key, expression_value):
        self._expression_key = expression_key
        self._expression_value = expression_value

    @property
    def expression_key(self):
        return self._expression_key

    @property
    def expression_value(self):
        return self._expression_value

    def evaluate(self, context):
        key = self.expression_key.evaluate(context)
        value = self.expression_value.evaluate(context)
        return (key, value)


# -----------------------------------------------------------------------------
# Iterable index
# -----------------------------------------------------------------------------

class IterableIndex(Atom):

    def __init__(self, name, expression):
        self._name = name
        self._expression = expression

    @property
    def name(self):
        return self._name

    @property
    def expression(self):
        return self._expression

    def evaluate(self, context):
        iterable = self.name.evaluate(context)
        index = self.expression.evaluate(context)
        return iterable[index]


# -----------------------------------------------------------------------------
# Function call
# -----------------------------------------------------------------------------

class FunctionCall(Atom):

    def __init__(self, name, expressions):
        self._name = name
        self._expressions = expressions

    @property
    def name(self):
        return self._name

    @property
    def expressions(self):
        return self._expressions
    
    def evaluate(self, context):
        identifier = self.name.identifier
        symbol = context.symtable.lookup(identifier)
        if (not symbol):
            raise RuntimeException(f'{identifier} 未定义')
        node = symbol.node

        # Check if calling a class constructor
        if (type(node) == ClassDef):
            return ASTUtil.instantiate_class(context, node)

        # Validate node type
        if (type(node) != FunctionDef):
            raise RuntimeException(f'{identifier} 不是一个函数')

        # Get function definition and its attributes
        functiondef = node
        parameters = functiondef.parameters
        body = functiondef.body

        # Create child context with a new symbol table
        function_symtable = SymbolTable(context.symtable)
        function_context = Context(function_symtable)

        # Insert symbols if default arguments exist
        if (functiondef.default_args):
            for parameter in parameters:
                arg_identifier = parameter.identifier
                if (arg_identifier in functiondef.default_args):
                    arg_value = functiondef.default_args[arg_identifier]
                    arg_node = Atom(arg_value)
                    function_context.symtable.insert(arg_identifier, arg_node)
                else:
                    continue
        
        # Assign expression values to parameters respectively
        for parameter, expression in zip(parameters, self.expressions):
            arg_identifier = parameter.identifier
            arg_node = Atom(expression.evaluate(context))
            function_context.symtable.insert(arg_identifier, arg_node)
        
        # Evaluate function body
        if (callable(body)):
            # Case: built-in function
            function_pointer = body
            args = dict()
            for parameter in parameters:
                value = parameter.evaluate(function_context)
                args[parameter.identifier] = value
            return_value = function_pointer(**args)
            return ASTUtil.evaluate_base_type(return_value)
        else:
            # Case: user-defined function
            value = ASTUtil.evaluate_block(
                function_context, body
            )
            if (type(value).__name__ == 'ReturnValue'):
                return value
            return ASTUtil.evaluate_base_type(None)


# -----------------------------------------------------------------------------
# Class member call
# -----------------------------------------------------------------------------

class ClassMemberCall(Atom):

    def __init__(self, name, members):
        self._name = name
        self._members = members
    
    @property
    def name(self):
        return self._name

    @property
    def members(self):
        return self._members

    def evaluate(self, context):
        instance = self.name.evaluate(context)
        value = instance
        for member in self.members:
            if (type(value) != ClassType):
                raise RuntimeException(
                    f'{self.name.identifier} 不是一个类实例'
                )
            value = member.evaluate(value.context)
        return value


# -----------------------------------------------------------------------------
# END ast.py
# -----------------------------------------------------------------------------



# -----------------------------------------------------------------------------
# sheng: utils.py
#
# Copyright (c) 2021
# luojiahai
# All rights reserved.
#
# Latest version: https://github.com/luojiahai/sheng
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -----------------------------------------------------------------------------


# -----------------------------------------------------------------------------
# BEGIN utils.py
# -----------------------------------------------------------------------------

import sys


class Global(object):

    class __Global(object):

        def __init__(self):
            self._is_debug = False
            self._input_stream = sys.stdin
            self._output_stream = sys.stdout
            self._log_stream = sys.stderr

        @property
        def input_stream(self):
            return self._input_stream

        @input_stream.setter
        def input_stream(self, other):
            self._input_stream = other
            
        @property
        def output_stream(self):
            return self._output_stream

        @output_stream.setter
        def output_stream(self, other):
            self._output_stream = other

        @property
        def log_stream(self):
            return self._log_stream

        @log_stream.setter
        def log_stream(self, other):
            self._log_stream = other

        @property
        def is_debug(self):
            return self._is_debug

        @is_debug.setter
        def is_debug(self, other):
            self._is_debug = other

        def log(self, message):
            print(message, file=Global().log_stream)

    __instance = None

    def __new__(cls):
        if (not Global.__instance):
            Global.__instance = Global.__Global()
        return Global.__instance


# -----------------------------------------------------------------------------
# END utils.py
# -----------------------------------------------------------------------------



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="sheng",
    version="0.1.19",
    author="luojiahai",
    author_email="luo@jiahai.co",
    description="The Sheng programming language: A Chinese programming language",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/luojiahai/sheng",
    project_urls={
        "Bug Tracker": "https://github.com/luojiahai/sheng/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "."},
    packages=setuptools.find_packages(where="."),
    python_requires=">=3.9",
    entry_points = {
        'console_scripts': ['sheng=src.driver:main'],
    }
)



# -----------------------------------------------------------------------------
# sheng: exception.py
#
# Copyright (c) 2021
# luojiahai
# All rights reserved.
#
# Latest version: https://github.com/luojiahai/sheng
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -----------------------------------------------------------------------------


# -----------------------------------------------------------------------------
# BEGIN exception.py
# -----------------------------------------------------------------------------

class ShengException(BaseException):

    def __init__(self, *args):
        super().__init__(*args)


class SyntaxException(ShengException):

    def __init__(self, message=''):
        super().__init__(f'「报错」【语法错误】 {message}')


class RuntimeException(ShengException):
    
    def __init__(self, message=''):
        super().__init__(f'「报错」【运行时错误】 {message}')
        

class AssertionException(ShengException):

    def __init__(self, message=''):
        super().__init__(f'「报错」【断言错误】 {message}')


# -----------------------------------------------------------------------------
# END exception.py
# -----------------------------------------------------------------------------



# -----------------------------------------------------------------------------
# sheng: lex.py
#
# Copyright (c) 2021
# luojiahai
# All rights reserved.
#
# Latest version: https://github.com/luojiahai/sheng
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -----------------------------------------------------------------------------


# -----------------------------------------------------------------------------
# BEGIN lex.py
# -----------------------------------------------------------------------------

from .ply import lex
from .builtins import *
from .exception import *
import itertools as _itertools


RESERVED_KEYWORDS = {
    '对': 'TRUE',
    '错': 'FALSE',
    '与': 'AND',
    '或': 'OR',
    '非': 'NOT',
    '开始': 'BEGIN',
    '结束': 'END',
    '赋值': 'EQUAL',
    '返回': 'RETURN',
    '略过': 'PASS',
    '断言': 'ASSERT',
    '停止': 'BREAK',
    '继续': 'CONTINUE',
    '定义': 'FUNCTIONDEF',
    '如果': 'IF',
    '则': 'THEN',
    '否则如果': 'ELIF',
    '否则': 'ELSE',
    '类': 'CLASSDEF',
    '迭代': 'ITERATE',
    '以': 'FOR',
    '循环': 'LOOP',
    '索引': 'INDEX',
    '空': 'NULL',
}

EXACT_TOKEN_TYPES = RESERVED_KEYWORDS | {
    ',': 'COMMA',
    '.': 'DOT',
    '(': 'LPAR',
    ')': 'RPAR',
    '[': 'LSQB',
    ']': 'RSQB',
    '{': 'LBRACE',
    '}': 'RBRACE',
    ':': 'COLON',
    '+': 'PLUS',
    '-': 'MINUS',
    '*': 'STAR',
    '/': 'SLASH',
    '**': 'DOUBLESTAR',
    '//': 'DOUBLESLASH',
    '%': 'PERCENT',
    '==': 'EQEQUAL',
    '!=': 'NOTEQUAL',
    '<': 'LESS',
    '<=': 'LESSEQUAL',
    '>': 'GREATER',
    '>=': 'GREATEREQUAL',
    '!': 'NOT',
}


def group(*choices): return '(' + '|'.join(choices) + ')'
def any(*choices): return group(*choices) + '*'
def maybe(*choices): return group(*choices) + '?'

# Ignore
Whitespace = r'[ \f\t]*'
Comment = r'\#[^\r\n]*'
Ignore = Whitespace + any(r'\\\r?\n' + Whitespace) + maybe(Comment)

# Number
Hexnumber = r'0[xX](?:_?[0-9a-fA-F])+'
Binnumber = r'0[bB](?:_?[01])+'
Octnumber = r'0[oO](?:_?[0-7])+'
Decnumber = r'(?:0(?:_?0)*|[1-9](?:_?[0-9])*)'
Intnumber = group(Hexnumber, Binnumber, Octnumber, Decnumber)
Exponent = r'[eE][-+]?[0-9](?:_?[0-9])*'
Pointfloat = group(r'[0-9](?:_?[0-9])*\.(?:[0-9](?:_?[0-9])*)?',
                r'\.[0-9](?:_?[0-9])*') + maybe(Exponent)
Expfloat = r'[0-9](?:_?[0-9])*' + Exponent
Floatnumber = group(Pointfloat, Expfloat)
Imagnumber = group(r'[0-9](?:_?[0-9])*[jJ]', Floatnumber + r'[jJ]')
Number = group(Imagnumber, Floatnumber, Intnumber)

# Return the empty string, plus all of the valid string prefixes.
def _all_string_prefixes():
    # The valid string prefixes. Only contain the lower case versions,
    #  and don't contain any permutations (include 'fr', but not
    #  'rf'). The various permutations will be generated.
    _valid_string_prefixes = ['b', 'r', 'u', 'f', 'br', 'fr']
    # if we add binary f-strings, add: ['fb', 'fbr']
    result = {''}
    for prefix in _valid_string_prefixes:
        for t in _itertools.permutations(prefix):
            # create a list with upper and lower versions of each
            #  character
            for u in _itertools.product(*[(c, c.upper()) for c in t]):
                result.add(''.join(u))
    return result

# Note that since _all_string_prefixes includes the empty string,
#  StringPrefix can be the empty string (making it optional).
StringPrefix = group(*_all_string_prefixes())

# Tail end of ' string.
Single = r"[^'\\]*(?:\\.[^'\\]*)*'"
# Tail end of " string.
Double = r'[^"\\]*(?:\\.[^"\\]*)*"'
# Tail end of ''' string.
Single3 = r"[^'\\]*(?:(?:\\.|'(?!''))[^'\\]*)*'''"
# Tail end of """ string.
Double3 = r'[^"\\]*(?:(?:\\.|"(?!""))[^"\\]*)*"""'
Triple = group(StringPrefix + "'''", StringPrefix + '"""')
# Single-line ' or " string.
String = group(StringPrefix + r"'[^\n'\\]*(?:\\.[^\n'\\]*)*'",
            StringPrefix + r'"[^\n"\\]*(?:\\.[^\n"\\]*)*"')


class Lexer(object):

    # List of token names
    tokens = tuple(set(EXACT_TOKEN_TYPES.values())) + (
        'INTEGER', 'FLOAT', 'COMPLEX', 'STRING', 'NAME',
    )

    # Regular expression rules for tokens
    t_COMMA = r','
    t_DOT = r'\.'
    t_LPAR = r'\('
    t_RPAR = r'\)'
    t_LSQB = r'\['
    t_RSQB = r'\]'
    t_LBRACE = r'{'
    t_RBRACE = r'}'
    t_COLON = r':'
    t_PLUS = r'\+'
    t_MINUS = r'-'
    t_STAR = r'\*'
    t_SLASH = r'/'
    t_DOUBLESTAR = r'\*\*'
    t_DOUBLESLASH = r'//'
    t_PERCENT = r'%'
    t_EQEQUAL = r'=='
    t_NOTEQUAL = r'!='
    t_LESS = r'<'
    t_LESSEQUAL = r'<='
    t_GREATER = r'>'
    t_GREATEREQUAL = r'>='
    t_TRUE = r'对'
    t_FALSE = r'错'
    t_NULL = r'空'
    t_AND = r'与'
    t_OR = r'或'
    t_NOT = r'非'
    t_BEGIN = r'开始'
    t_END = r'结束'
    t_EQUAL = r'赋值'
    t_RETURN = r'返回'
    t_PASS = r'略过'
    t_ASSERT = r'断言'
    t_BREAK = r'停止'
    t_CONTINUE = r'继续'
    t_FUNCTIONDEF = r'定义'
    t_IF = r'如果'
    t_THEN = r'则'
    t_ELIF = r'否则如果'
    t_ELSE = r'否则'
    t_CLASSDEF = r'类'
    t_ITERATE = r'迭代'
    t_FOR = r'以'
    t_LOOP = r'循环'
    t_INDEX = r'索引'
    t_ignore_COMMENT = Comment
    t_ignore = ' \t'
    t_INTEGER = Intnumber
    t_FLOAT = Floatnumber
    t_COMPLEX = Imagnumber
    t_STRING = String

    # Define identifier rule that can only allow chinese characters
    def t_NAME(self, t):
        r'[^\x00-\xff]+'
        if (t.value in RESERVED_KEYWORDS):
            t.type = RESERVED_KEYWORDS[t.value]
        return t

    # Define a rule so we can track line numbers
    def t_NEWLINE(self, t):
        r'\n'
        t.lexer.lineno += 1

    # Error handling rule
    def t_error(self, t):
        raise SyntaxException(f'无效的字符 \'{t.value[0]}\'，第 {t.lexer.lineno} 行')
        # t.lexer.skip(1)

    # Build the lexer
    def build(self, **kwargs):
        self.lexer = lex.lex(module=self, **kwargs)


# -----------------------------------------------------------------------------
# END lex.py
# -----------------------------------------------------------------------------



# -----------------------------------------------------------------------------
# BEGIN driver.py
# -----------------------------------------------------------------------------

import os
import sys
import time
from optparse import OptionParser
from src.sheng.compile import *
from src.sheng.utils import *


LOG_DIRECTORY = './log/'

SYSTEM_STREAMS = {
    'stdin': sys.stdin,
    'stdout': sys.stdout,
    'stderr': sys.stderr,
}


def default(str):
    return str + ' [Default: %default]'

def read_command(argv):
    usage_str = 'sheng [option] [file]'
    parser = OptionParser(usage_str)

    parser.add_option(
        '-d', '--debug', action='store_true', dest='debug',
        help=default('debug mode'), default=False
    )
    
    parser.add_option(
        '-i', '--input-stream', dest='input_stream',
        help=default('input stream'), default='stdin'
    )

    parser.add_option(
        '-o', '--output-stream', dest='output_stream',
        help=default('output stream'), default='stdout'
    )

    parser.add_option(
        '-l', '--log-stream', dest='log_stream',
        help=default('log stream, applicable only if debug mode is on'), 
        default=f'sheng_{time.strftime("%Y-%m-%d-%H-%M-%S", time.localtime())}.log'
    )

    options, other = parser.parse_args(argv[1:])
    if (len(other) == 0):
        parser.error('no input file')
    if (len(other) > 1):
        parser.error('Command line input not understood: ' + str(other[1:]))

    args = dict()
    args['file'] = other[0]

    # Debug, write logs to log stream
    if (options.debug):
        Global().is_debug = options.debug
        log_stream = None
        if (options.log_stream in SYSTEM_STREAMS):
            log_stream = SYSTEM_STREAMS[options.log_stream]
        else:
            log_stream_path = LOG_DIRECTORY + options.log_stream
            os.makedirs(os.path.dirname(log_stream_path), exist_ok=True)
            log_stream = open(log_stream_path, 'a')
        Global().log_stream = log_stream

    # Input stream
    input_stream = None
    if (options.input_stream in SYSTEM_STREAMS):
        input_stream = SYSTEM_STREAMS[options.input_stream]
    else:
        input_stream = open(options.input_stream, 'r')
    Global().input_stream = input_stream

    # Output stream
    output_stream = None
    if (options.output_stream in SYSTEM_STREAMS):
        output_stream = SYSTEM_STREAMS[options.output_stream]
    else:
        output_stream = open(options.output_stream, 'w')
    Global().output_stream = output_stream

    return args

def read_file(file):
    f = open(file, "r", encoding='utf-8')
    data = f.read()
    f.close()
    return data

def main():
    args = read_command(sys.argv)
    data = read_file(**args)
    return execute(data)


# -----------------------------------------------------------------------------
# END driver.py
# -----------------------------------------------------------------------------
